// Modified H-basis and Irradiance Parallax Occlusion Mapping

// Parallax Occlusion Mapping part is from AMD Render Monkey examples

//Global compress range is set by hand or defined by the maximum value in the textures [texturename]_1..5
//See baking script for compression.
#define COMPRESSRANGE 0.75 //set to 1 for a compress range of -1..1

//Global multipliers
#define GLOBAL_SCALE 2.5 //2.5066282 is sqrt2pi
#define GLOBAL_TEXTURE 2.5

#define OVERSQ2PI 0.39894228		
#define SQ32PI 0.69098830		
#define	SQ152PI	1.5450968

// Irradiance Parralax Occlusion Mapping parameters
#define fHeightMapRange 0.02
#define fShadowSoftening 0.3

#define nMinSamples 12
#define nMaxSamples 50

void IpomVP(    float4 position	: POSITION,
				float3 normal : NORMAL,
				float3 tangent : TANGENT,
				float3 binormal : BINORMAL,
				float2 normalUV	: TEXCOORD0,
				float2 lightUV	: TEXCOORD1,
										
				out float4 oPosition : POSITION,
				out float2 onormalUV : TEXCOORD0,
				out float2 olightUV : TEXCOORD1,
				out float3 vViewTS  : TEXCOORD2,   // view vector in tangent space, denormalized
				out	float2 vParallaxOffsetTS : TEXCOORD3,   // Parallax offset vector in tangent space
				out	float3 vNormalWS : TEXCOORD4,   // Normal vector in world space
				out	float3 vViewWS : TEXCOORD5,   // View vector in world space
				
				uniform float4x4 worldViewProj,
				uniform float4x4 viewInverse
            )

{	
	onormalUV = normalUV;
	olightUV = lightUV;
	vNormalWS = normal;
	
   // Compute and output the world view vector:
   float3 view = mul( viewInverse, float4( 0, 0, 0, 1)) - position;
   vViewWS = view;
   // Compute denormalized light vector in world space:
   // float3 vLightWS = vLightPosition - position;
   oPosition = mul(worldViewProj, position);
    // Normalize the light and view vectors and transform it to the tangent space:
   float3x3 mWorldToTangent = float3x3( tangent, binormal, normal );
   
   // Propagate the view vector (in tangent space):
   vViewTS  = mul( mWorldToTangent, vViewWS );
   
   // Compute initial parallax displacement direction:
   float2 vParallaxDirection = normalize( vViewTS.xy );
   
   // The length of this vector determines the furthest amount of displacement:
   float fLength = length( vViewTS );
   float fParallaxLength = sqrt( fLength * fLength - vViewTS.z * vViewTS.z ) / vViewTS.z;
   
   // Compute the actual reverse parallax displacement vector:
   vParallaxOffsetTS = vParallaxDirection * fParallaxLength;
   
   // Need to scale the amount of displacement to account for different height ranges
   // in height maps. This is controlled by an artist-editable parameter:
   vParallaxOffsetTS *= fHeightMapRange;
}

void HmodNormVP(float4 position	: POSITION,
				float2 normalUV	: TEXCOORD0,
				float2 lightUV	: TEXCOORD1,
				
				
							
				out float4 oPosition : POSITION,
				out float2 onormalUV : TEXCOORD0,
				out float2 olightUV : TEXCOORD1,
				
				uniform float4x4 worldViewProj

            )

{	
	oPosition = mul(worldViewProj, position);
	
	onormalUV = normalUV;
	olightUV = lightUV;

}
//------------------------------------------------------------------------------
void IpomFP(	float2 normalUV : TEXCOORD0,
				float2 lightUV  : TEXCOORD1,
				float3 vViewTS  : TEXCOORD2,   // view vector in tangent space, denormalized
				float2 vParallaxOffsetTS : TEXCOORD3,   // Parallax offset vector in tangent space
				float3 vNormalWS : TEXCOORD4,   // Normal vector in world space
				float3 vViewWS : TEXCOORD5,   // View vector in world space
				
				
				out float4 color	: COLOR,
		 
				uniform sampler2D albedo : register(s0),
				uniform sampler2D normal : register(s1),
				uniform sampler2D height : register(s2),
				uniform sampler2D coeff0 : register(s3),
				uniform sampler2D coeff1 : register(s4),
				uniform sampler2D coeff2 : register(s5),
				uniform sampler2D coeff3 : register(s6),
				uniform sampler2D coeff4 : register(s7),
				uniform sampler2D coeff5 : register(s8)
)
{
		//  Normalize the interpolated vectors:
		vViewTS   = normalize( vViewTS  );
		vViewWS   = normalize( vViewWS  );
		//vLightTS  = normalize( vLightTS );
		vNormalWS = normalize( vNormalWS );

		// Compute all the derivatives:
		float2 dx = ddx( normalUV );
		float2 dy = ddy( normalUV );

	    ////===============================================//
	    //// Parallax occlusion mapping offset computation //
	    ////===============================================//

	    //// Utilize dynamic flow control to change the number of samples per ray 
	    //// depending on the viewing angle for the surface. Oblique angles require 
	    //// smaller step sizes to achieve more accurate precision for computing displacement.
	    //// We express the sampling rate as a linear function of the angle between 
		//// the geometric normal and the view direction ray:
		int nNumSteps = (int) lerp( nMaxSamples, nMinSamples, dot( vViewWS, vNormalWS ) );
		//// Intersect the view ray with the height field profile along the direction of
		//// the parallax offset ray (computed in the vertex shader. Note that the code is
		//// designed specifically to take advantage of the dynamic flow control constructs
	    //// in HLSL and is very sensitive to specific syntax. When converting to other examples,
	    //// if still want to use dynamic flow control in the resulting assembly shader,
	    //// care must be applied.
	    //// 
	    //// In the below steps we approximate the height field profile as piecewise linear
	    //// curve. We find the pair of endpoints between which the intersection between the 
	    //// height field profile and the view ray is found and then compute line segment
	    //// intersection for the view ray and the line segment formed by the two endpoints.
	    //// This intersection is the displacement offset from the original texture coordinate.
	    //// See the above paper for more details about the process and derivation.
	    ////
		//
	    float fCurrHeight = 0.0;
	    float fStepSize   = 1.0 / (float) nNumSteps;
	    float fPrevHeight = 1.0;
	    float fNextHeight = 0.0;

	    int    nStepIndex = 0;
	    bool   bCondition = true;

	    float2 vTexOffsetPerStep = fStepSize * vParallaxOffsetTS;
	    float2 vTexCurrentOffset = normalUV;
	    float  fCurrentBound     = 1.0;
	    float  fParallaxAmount   = 0.0;

	    float2 pt1 = 0;
	    float2 pt2 = 0;

	    float2 texOffset2 = 0;
		
		while ( nStepIndex < nNumSteps ) 
		{
		  vTexCurrentOffset -= vTexOffsetPerStep;

		  // Sample height map which in this case is stored in the alpha channel of the normal map:
		  fCurrHeight = tex2Dgrad( height, vTexCurrentOffset, dx, dy ).r;

		  fCurrentBound -= fStepSize;

		  if ( fCurrHeight > fCurrentBound ) 
		  {     
			 pt1 = float2( fCurrentBound, fCurrHeight );
			 pt2 = float2( fCurrentBound + fStepSize, fPrevHeight );

			 texOffset2 = vTexCurrentOffset - vTexOffsetPerStep;

			 nStepIndex = nNumSteps + 1;
		  }
		  else
		  {
			 nStepIndex++;
			 fPrevHeight = fCurrHeight;
		  }
		}   // End of while ( nStepIndex < nNumSteps )
		//
	   float fDelta2 = pt2.x - pt2.y;
	   float fDelta1 = pt1.x - pt1.y;
	   fParallaxAmount = (pt1.x * fDelta2 - pt2.x * fDelta1 ) / ( fDelta2 - fDelta1 );
	   
	   float2 vParallaxOffset = vParallaxOffsetTS * (1 - fParallaxAmount );

	   // The computed texture offset for the displaced point on the pseudo-extruded surface:
	   float2 texSample = normalUV - vParallaxOffset;
	     
	    // Irradiance calculation + albedo
	   float4 alb = pow(tex2D(albedo,texSample),2.2); //srgb to linear, sRGB state does not work in Ogre so we do our own conversion
	   float3 norm = 2*tex2D(normal,texSample)-1;
		
	   float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI //this should be done with sRGB lookup
					+(2*COMPRESSRANGE*tex2D(coeff1,lightUV)-COMPRESSRANGE)*SQ32PI*norm.y
					+(2*COMPRESSRANGE*tex2D(coeff2,lightUV)-COMPRESSRANGE)*SQ32PI*(1-norm.z)
					+(2*COMPRESSRANGE*tex2D(coeff3,lightUV)-COMPRESSRANGE)*SQ32PI*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff4,lightUV)-COMPRESSRANGE)*SQ152PI*norm.y*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff5,lightUV)-COMPRESSRANGE)*SQ152PI*0.5*(norm.x*norm.x -norm.y*norm.y); //0.5 as in equation
										
	   color.rgb = GLOBAL_SCALE*Irr*alb.rgb*GLOBAL_TEXTURE;
	   color.a = 1.0;				
}
void IpomNoAlbedoFP(	float2 normalUV : TEXCOORD0,
				float2 lightUV  : TEXCOORD1,
				float3 vViewTS  : TEXCOORD2,   // view vector in tangent space, denormalized
				float2 vParallaxOffsetTS : TEXCOORD3,   // Parallax offset vector in tangent space
				float3 vNormalWS : TEXCOORD4,   // Normal vector in world space
				float3 vViewWS : TEXCOORD5,   // View vector in world space
				
				
				out float4 color	: COLOR,
		 
				uniform sampler2D normal : register(s0),
				uniform sampler2D height : register(s1),
				uniform sampler2D coeff0 : register(s2),
				uniform sampler2D coeff1 : register(s3),
				uniform sampler2D coeff2 : register(s4),
				uniform sampler2D coeff3 : register(s5),
				uniform sampler2D coeff4 : register(s6),
				uniform sampler2D coeff5 : register(s7)
)
{
		//  Normalize the interpolated vectors:
		vViewTS   = normalize( vViewTS  );
		vViewWS   = normalize( vViewWS  );
		//vLightTS  = normalize( vLightTS );
		vNormalWS = normalize( vNormalWS );
	     
		//float4 cResultColor = float4( 0, 0, 0, 1 );

		// Compute all the derivatives:
		float2 dx = ddx( normalUV );
		float2 dy = ddy( normalUV );

	    ////===============================================//
	    //// Parallax occlusion mapping offset computation //
	    ////===============================================//

	    //// Utilize dynamic flow control to change the number of samples per ray 
	    //// depending on the viewing angle for the surface. Oblique angles require 
	    //// smaller step sizes to achieve more accurate precision for computing displacement.
	    //// We express the sampling rate as a linear function of the angle between 
		//// the geometric normal and the view direction ray:
		int nNumSteps = (int) lerp( nMaxSamples, nMinSamples, dot( vViewWS, vNormalWS ) );
		//// Intersect the view ray with the height field profile along the direction of
		//// the parallax offset ray (computed in the vertex shader. Note that the code is
		//// designed specifically to take advantage of the dynamic flow control constructs
	    //// in HLSL and is very sensitive to specific syntax. When converting to other examples,
	    //// if still want to use dynamic flow control in the resulting assembly shader,
	    //// care must be applied.
	    //// 
	    //// In the below steps we approximate the height field profile as piecewise linear
	    //// curve. We find the pair of endpoints between which the intersection between the 
	    //// height field profile and the view ray is found and then compute line segment
	    //// intersection for the view ray and the line segment formed by the two endpoints.
	    //// This intersection is the displacement offset from the original texture coordinate.
	    //// See the above paper for more details about the process and derivation.
	    ////
		//
	    float fCurrHeight = 0.0;
	    float fStepSize   = 1.0 / (float) nNumSteps;
	    float fPrevHeight = 1.0;
	    float fNextHeight = 0.0;

	    int    nStepIndex = 0;
	    bool   bCondition = true;

	    float2 vTexOffsetPerStep = fStepSize * vParallaxOffsetTS;
	    float2 vTexCurrentOffset = normalUV;
	    float  fCurrentBound     = 1.0;
	    float  fParallaxAmount   = 0.0;

	    float2 pt1 = 0;
	    float2 pt2 = 0;

	    float2 texOffset2 = 0;
		
		while ( nStepIndex < nNumSteps ) 
		{
		  vTexCurrentOffset -= vTexOffsetPerStep;

		  // Sample height map which in this case is stored in the alpha channel of the normal map:
		  fCurrHeight = tex2Dgrad( height, vTexCurrentOffset, dx, dy ).r;

		  fCurrentBound -= fStepSize;

		  if ( fCurrHeight > fCurrentBound ) 
		  {     
			 pt1 = float2( fCurrentBound, fCurrHeight );
			 pt2 = float2( fCurrentBound + fStepSize, fPrevHeight );

			 texOffset2 = vTexCurrentOffset - vTexOffsetPerStep;

			 nStepIndex = nNumSteps + 1;
		  }
		  else
		  {
			 nStepIndex++;
			 fPrevHeight = fCurrHeight;
		  }
		}   // End of while ( nStepIndex < nNumSteps )
	
	   float fDelta2 = pt2.x - pt2.y;
	   float fDelta1 = pt1.x - pt1.y;
	   fParallaxAmount = (pt1.x * fDelta2 - pt2.x * fDelta1 ) / ( fDelta2 - fDelta1 );
	   
	   float2 vParallaxOffset = vParallaxOffsetTS * (1 - fParallaxAmount );

	   // The computed texture offset for the displaced point on the pseudo-extruded surface:
	   float2 texSample = normalUV - vParallaxOffset;
	  
	   float3 norm = 2*tex2D(normal,texSample)-1;
		
	   float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI //this should be done with sRGB lookup
					+(2*COMPRESSRANGE*tex2D(coeff1,lightUV)-COMPRESSRANGE)*SQ32PI*norm.y
					+(2*COMPRESSRANGE*tex2D(coeff2,lightUV)-COMPRESSRANGE)*SQ32PI*(1-norm.z)
					+(2*COMPRESSRANGE*tex2D(coeff3,lightUV)-COMPRESSRANGE)*SQ32PI*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff4,lightUV)-COMPRESSRANGE)*SQ152PI*norm.y*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff5,lightUV)-COMPRESSRANGE)*SQ152PI*0.5*(norm.x*norm.x -norm.y*norm.y); //0.5 as in equation
										
		color.rgb = GLOBAL_SCALE*Irr;
		color.a = 1.0;				
}
void HmodNormFP(float2 normalUV      : TEXCOORD0,
				float2 lightUV       : TEXCOORD1,
				
				out float4 color	: COLOR,
		 
				uniform sampler2D albedo : register(s0),
				uniform sampler2D normal : register(s1),
				uniform sampler2D coeff0 : register(s2),
				uniform sampler2D coeff1 : register(s3),
				uniform sampler2D coeff2 : register(s4),
				uniform sampler2D coeff3 : register(s5),
				uniform sampler2D coeff4 : register(s6),
				uniform sampler2D coeff5 : register(s7)

)
{
		float4 alb = pow(tex2D(albedo,normalUV),2.2); //srgb to linear, sRGB state does not work in Ogre so we do our own conversion
		float3 norm = 2*tex2D(normal,normalUV)-1;
		
		float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI //this should be done with sRGB lookup
					+(2*COMPRESSRANGE*tex2D(coeff1,lightUV)-COMPRESSRANGE)*SQ32PI*norm.y
					+(2*COMPRESSRANGE*tex2D(coeff2,lightUV)-COMPRESSRANGE)*SQ32PI*(1-norm.z)
					+(2*COMPRESSRANGE*tex2D(coeff3,lightUV)-COMPRESSRANGE)*SQ32PI*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff4,lightUV)-COMPRESSRANGE)*SQ152PI*norm.y*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff5,lightUV)-COMPRESSRANGE)*SQ152PI*0.5*(norm.x*norm.x -norm.y*norm.y); //0.5 as in equation
										
		color.rgb = GLOBAL_SCALE*Irr*alb.rgb*GLOBAL_TEXTURE;
		color.a = 1.0;				
}



void HmodNormNoAlbedoFP(float2 normalUV      : TEXCOORD0,
				float2 lightUV       : TEXCOORD1,
				
				out float4 color	: COLOR,
		 
				uniform sampler2D normal : register(s0),
				uniform sampler2D coeff0 : register(s1),
				uniform sampler2D coeff1 : register(s2),
				uniform sampler2D coeff2 : register(s3),
				uniform sampler2D coeff3 : register(s4),
				uniform sampler2D coeff4 : register(s5),
				uniform sampler2D coeff5 : register(s6)

)
{
		float3 norm = 2*tex2D(normal,normalUV)-1;
		
		float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI 
					+(2*COMPRESSRANGE*tex2D(coeff1,lightUV)-COMPRESSRANGE)*SQ32PI*norm.y
					+(2*COMPRESSRANGE*tex2D(coeff2,lightUV)-COMPRESSRANGE)*SQ32PI*(1-norm.z)
					+(2*COMPRESSRANGE*tex2D(coeff3,lightUV)-COMPRESSRANGE)*SQ32PI*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff4,lightUV)-COMPRESSRANGE)*SQ152PI*norm.y*norm.x
					+(2*COMPRESSRANGE*tex2D(coeff5,lightUV)-COMPRESSRANGE)*SQ152PI*0.5*(norm.x*norm.x -norm.y*norm.y); //0.5 as in equation
										
		color.rgb = GLOBAL_SCALE*Irr;
		color.a = 1.0;				
}

void HmodNormLinFP(float2 normalUV      : TEXCOORD0,
				float2 lightUV       : TEXCOORD1,
				
				out float4 color	: COLOR,
		 
				uniform sampler2D albedo : register(s0),
				uniform sampler2D normal : register(s1),
				uniform sampler2D coeff0 : register(s2),
				uniform sampler2D coeff1 : register(s3),
				uniform sampler2D coeff2 : register(s4),
				uniform sampler2D coeff3 : register(s5)
)
{
		float4 alb = pow(tex2D(albedo,normalUV),2.2); //we do our own sRGB to linear, sRGB Gamma correction in the renderwindow must be set.
		float3 norm = 2*tex2D(normal,normalUV)-1;

		float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI 
					+(2*COMPRESSRANGE*tex2D(coeff1,lightUV)-COMPRESSRANGE)*SQ32PI*norm.y
					+(2*COMPRESSRANGE*tex2D(coeff2,lightUV)-COMPRESSRANGE)*SQ32PI*(1-norm.z)
					+(2*COMPRESSRANGE*tex2D(coeff3,lightUV)-COMPRESSRANGE)*SQ32PI*norm.x;
										
		color.rgb = GLOBAL_SCALE*Irr*alb.rgb*GLOBAL_TEXTURE;	
		color.a = 1.0;			
}

void HmodNormLinNoAlbedoFP(float2 normalUV      : TEXCOORD0,
				float2 lightUV       : TEXCOORD1,
				
				out float4 color	: COLOR,
		 
				uniform sampler2D normal : register(s0),
				uniform sampler2D coeff0 : register(s1),
				uniform sampler2D coeff1 : register(s2),
				uniform sampler2D coeff2 : register(s3),
				uniform sampler2D coeff3 : register(s4)
)
{
		float3 norm = 2*tex2D(normal,normalUV)-1;
		
		float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI 
					+(2*COMPRESSRANGE*tex2D(coeff1,lightUV)-COMPRESSRANGE)*SQ32PI*norm.y
					+(2*COMPRESSRANGE*tex2D(coeff2,lightUV)-COMPRESSRANGE)*SQ32PI*(1-norm.z)
					+(2*COMPRESSRANGE*tex2D(coeff3,lightUV)-COMPRESSRANGE)*SQ32PI*norm.x;

										
		color.rgb = GLOBAL_SCALE*Irr;
		color.a = 1.0;		
}

void HmodNormConstFP(float2 normalUV      : TEXCOORD0,
				float2 lightUV       : TEXCOORD1,
				
				out float4 color	: COLOR,
		 
				uniform sampler2D albedo : register(s0),
				uniform sampler2D coeff0 : register(s1)
)
{
		float4 alb = pow(tex2D(albedo,normalUV),2.2); //we do our own sRGB to linear, sRGB Gamma correction in the renderwindow must be set.

		float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI;
					
		color.rgb = GLOBAL_SCALE*Irr*alb.rgb*GLOBAL_TEXTURE;
		color.a = 1.0;			
}

void HmodNormConstNoAlbedoFP(float2 normalUV      : TEXCOORD0,
				float2 lightUV       : TEXCOORD1,
				
				out float4 color	: COLOR,
				
				uniform sampler2D coeff0 : register(s0)
)
{

		float3 Irr = pow(tex2D(coeff0,lightUV),2.2)*OVERSQ2PI;
									
		color.rgb =GLOBAL_SCALE*Irr;
		color.a = 1.0;		
}



